from ..config import permissions, validate_ip
from ..exceptions import CyberarkAPIException, CyberarkException, UserAlreadyMember, UserNotExists, SafeAlreadyExistsError
from ..cyberark import EPV
import json
import re


class PrivilegedAccount:
    def __init__(self, hostname: str, address: str, username: str, password: str, safe: str, platform: str,
                 fonction: str, metier: str, environment: str):
        self.hostname = hostname
        self.address = address
        self.username = username
        self.platform = platform
        self.password = password
        self.safe = safe
        self.id = ""
        self.fonction = fonction
        self.metier = metier
        self.environment = environment

    def get_name(self):
        return "{}-{}".format(self.address, self.username)

    def to_json(self):
        json_object = {
            "name": self.get_name(),
            "address": self.hostname,
            "userName": self.username,
            "platformId": self.platform,
            "safeName": self.safe,
            "secret": self.password,
            "platformAccountProperties": {
                "Fonction": self.fonction,
                "Metier": self.metier,
                "IP": self.address,
                "IPCible": self.address,
                "Realm": self.environment,
                "CreationMethod": "RestAPI"
            },
            "secretManagement": {
                "automaticManagementEnabled": False,  # automatically managed by the CPM
                "manualManagementReason": "Unmanaged team acccount"
            }
        }

        return json_object

    def __str__(self):
        strrepr = self.to_json()
        strrepr["secret"] = "***********"
        return str(strrepr)


class KeeAccount(PrivilegedAccount):
    def __init__(self, hostname: str, address: str, username: str, password: str, safe: str, platform: str,
                 fonction: str, metier: str, environment: str, database: str):

        super().__init__(hostname, address, username, password, safe, platform, fonction, metier, environment)
        self.AccountType = "Keepass"
        self.database = database


    def to_json(self):
        json_object = super(KeeAccount, self).to_json()
        json_object["platformAccountProperties"]["AccountType"] = self.AccountType
        json_object["platformAccountProperties"]["Database"] = self.database
        return json_object


class Account:
    def __init__(self, epv: EPV):
        self.epv = epv

    def add_account_to_safe(self, account: PrivilegedAccount):
        url = 'https://' + self.epv.api_host + '/PasswordVault/api/Accounts'
        head = {"Content-type": "application/json", "Authorization": self.epv.token}

        try:
            req = self.epv.session.post(url, headers=head, data=json.dumps(account.to_json()), timeout=30,
                                      verify=False)
        except Exception as err:
            # uncatched exception
            raise CyberarkException(err)

        if req.status_code != 201:
            content = json.loads(req.content)
            raise CyberarkAPIException(content["ErrorMessage"])

        acc_id = json.loads(req.content)['id']
        account.id = acc_id


    def is_valid_username(name: str) -> bool:
        return re.match("[^@]+@[^@]+[.][^@]+", name) is not None


    def is_valid_safename(name: str) -> bool:
        return re.match("^[0-9]{12}$", name) is not None


    def user_exists(self, username: str):
        if self.epv.user_list is None:
            url = 'https://' + self.epv.api_host + '/PasswordVault/api/Users/'
            head = {"Content-type": "application/json", "Authorization": self.epv.token}
            req = self.epv.session.get(url, headers=head, timeout=30, verify=False)
            if req.status_code != 200:
                raise Exception("Error occured during user_exists call. CyberArk error:", req.content)
            self.epv.user_list = [u['username'].lower().strip() for u in json.loads(req.content)['Users']]
        return username in self.epv.user_list


    def add_safe_member(self, safe: str, username: str, profile: str):
        url = 'https://' + self.epv.api_host + '/PasswordVault/WebServices/PIMServices.svc/Safes/' + safe + '/Members'
        head = {"Content-type": "application/json", "Authorization": self.epv.token}
        data = {'member': {
            'MemberName': username,
            'Permissions': permissions(profile)
        }
        }
        req = self.epv.session.post(url, headers=head, data=json.dumps(data), timeout=30, verify=False)
        if req.status_code != 201:
            content = json.loads(req.content)
            if "Reason: User or Group is already a member." in content["ErrorMessage"]:
                raise UserAlreadyMember
            elif "Reason: Input parameter for [MemberName] value is invalid" in content["ErrorMessage"]:
                raise UserNotExists
            else:
                raise Exception("Error occured during add_safe_member call. CyberArk error:", req.content)


    def remove_safe_member(self, safe: str, username: str):
        url = 'https://' + self.epv.api_host + '/PasswordVault/WebServices/PIMServices.svc/Safes/' + safe + \
              '/Members/' + username
        head = {"Content-type": "application/json", "Authorization": self.epv.token}
        req = self.epv.session.delete(url, headers=head, timeout=30, verify=False)
        if req.status_code != 200:
            raise Exception("Error occured during remove_safe_member call. CyberArk error:", req.content)


    def safe_exists(self, safename: str):
        url = 'https://' + self.epv.api_host + '/PasswordVault/WebServices/PIMServices.svc/Safes/?query=' + safename
        head = {"Content-type": "application/json", "Authorization": self.epv.token}
        req = self.epv.session.get(url, headers=head, timeout=30, verify=False)
        if req.status_code == 200:
            return True
        elif req.status_code == 404:
            return False
        else:
            raise Exception("Error occured during remove_safe_member call. CyberArk error:", req.content)

    def create_safe(self, safename: str, description=""):
        url = "https://" + self.epv.api_host + "/PasswordVault/WebServices/PIMServices.svc/Safes"
        data = {
            "safe": {
                "SafeName": safename,
                "Description": description,
                "OLACEnabled": False,
                "ManagingCPM": self.epv.cpm,
                "NumberOfVersionsRetention": self.epv.retention
            }
        }
        headers = {"Content-type": "application/json", "Authorization": self.epv.token}

        try:
            req = self.epv.session.post(url=url, data=json.dumps(data), headers=headers, verify=False)
        except Exception:
            raise CyberarkException

        if req.status_code != 201:
            error_code = req.json()["ErrorCode"]
            if error_code == 'CAWS00001E':
                raise SafeAlreadyExistsError()
            else:
                raise CyberarkException("Error occured during add_safe call. CyberArk error:")
        return 0


    def add_safe(self, safename: str, description=""):
        try:
            self.create_safe(safename, description)
        except Exception as e:
            raise e

        # Define Safe defaults owners
        for user, profile in {"Vault Admins": "admin", "onboarding": "prov", "Administrator": 'admin'}.items():
            try:
                self.add_safe_member(safename, user, profile)
            except UserAlreadyMember:
                pass
        # when we did that, we had a bug so we remove it
        # try:
        #     # Removing provisioning user from admin of the safe
        #     self.remove_safe_member(safename, config.username)
        # except CyberarkException:
        #     raise CyberarkException("Unable to remove %s from safe", config.username)


    def search_account(self, account: (PrivilegedAccount, str)) -> list:
        if type(account) is str:
            address = account
        elif isinstance(account, PrivilegedAccount):
            address = account.address
        else:
            raise TypeError("search account function returned : First argument of account is not valid")

        if not validate_ip(address):
            raise TypeError("The address of the object was not considered as valid IPv4 address %s", address)

        url = 'https://' + self.epv.api_host + '/PasswordVault/api/Accounts?search=' + address
        if isinstance(account, PrivilegedAccount):
            # Searching for both username + IP Address
            url += " " + account.username + "&limit=111"
        else:
            url += "&limit=111"

        headers = {"Content-type": "application/json", "Authorization": self.epv.token}

        req = self.epv.session.get(url=url, headers=headers, verify=False)

        if req.status_code != 200:
            raise Exception("Error occured during search_account call. CyberArk error:")

        json_response_data = json.loads(req.content)
        exact_matches = []
        for result in json_response_data['value']:
            if result['platformAccountProperties']['IP'] == address:
                exact_matches.append(result)
        return exact_matches


# Patch Top level Object
EPV.Account = Account
